import numpy as np
from luvina.utils.data_utils import get_file


file_name = "wikipedia_common_misspellings.txt"
origin = ('https://raw.githubusercontent.com/rameshjesswani/Semantic-Textual' +
          '-Similarity/master/nlp_basics/nltk/dataset/dataset_misspelled.txt')


def load_data():
    """
    i. Data set is downloaded from:
    https://en.wikipedia.org/wiki/Wikipedia:Lists_of_common_misspellings/A
    ii. Data set contains misspelled words and correct words
    iii. Words are selected randomly from data set
    iv. Around 670 words are selected randomly from data set to evaluate
    the performance of spell correctors
    """
    file_path = get_file(file_name, origin)
    load_file = np.loadtxt(file_path, dtype='str')

    return load_file



import enchant

# -------------------------------------------------------------------
# ENCHANT GLOBAL VARIABLES
# -------------------------------------------------------------------
spell_dictionary = enchant.Dict('en')
# -------------------------------------------------------------------


def in_dictionary(token):
    """ checks if a token is found in the dictionary or not
    args:
        token: string
    returns:
        boolean determining if the token was found in dictionary
    """
    return spell_dictionary.check(token)


def suggest_words(token):
    """ suggest similar tokens
    args:
        token: string
    returns:
        list of strings containing similar tokens
    """
    return spell_dictionary.suggest(token)



""" File taken directly from the Keras repo
Python utilities required by luvina"""
from __future__ import absolute_import

import numpy as np

import time
import sys


class Progbar(object):
    """Displays a progress bar.

    # Arguments
        target: Total number of steps expected, None if unknown.
        interval: Minimum visual progress update interval (in seconds).
    """

    def __init__(self, target, width=30, verbose=1, interval=0.05):
        self.width = width
        if target is None:
            target = -1
        self.target = target
        self.sum_values = {}
        self.unique_values = []
        self.start = time.time()
        self.last_update = 0
        self.interval = interval
        self.total_width = 0
        self.seen_so_far = 0
        self.verbose = verbose

    def update(self, current, values=None, force=False):
        """Updates the progress bar.

        # Arguments
            current: Index of current step.
            values: List of tuples (name, value_for_last_step).
                The progress bar will display averages for these values.
            force: Whether to force visual progress update.
        """
        values = values or []
        for k, v in values:
            if k not in self.sum_values:
                self.sum_values[k] = [v * (current - self.seen_so_far),
                                      current - self.seen_so_far]
                self.unique_values.append(k)
            else:
                self.sum_values[k][0] += v * (current - self.seen_so_far)
                self.sum_values[k][1] += (current - self.seen_so_far)
        self.seen_so_far = current

        now = time.time()
        if self.verbose == 1:
            if not force and (now - self.last_update) < self.interval:
                return

            prev_total_width = self.total_width
            sys.stdout.write('\b' * prev_total_width)
            sys.stdout.write('\r')

            if self.target is not -1:
                numdigits = int(np.floor(np.log10(self.target))) + 1
                barstr = '%%%dd/%%%dd [' % (numdigits, numdigits)
                bar = barstr % (current, self.target)
                prog = float(current) / self.target
                prog_width = int(self.width * prog)
                if prog_width > 0:
                    bar += ('=' * (prog_width - 1))
                    if current < self.target:
                        bar += '>'
                    else:
                        bar += '='
                bar += ('.' * (self.width - prog_width))
                bar += ']'
                sys.stdout.write(bar)
                self.total_width = len(bar)

            if current:
                time_per_unit = (now - self.start) / current
            else:
                time_per_unit = 0
            eta = time_per_unit * (self.target - current)
            info = ''
            if current < self.target and self.target is not -1:
                info += ' - ETA: %ds' % eta
            else:
                info += ' - %ds' % (now - self.start)
            for k in self.unique_values:
                info += ' - %s:' % k
                if isinstance(self.sum_values[k], list):
                    avg = self.sum_values[k][0] / max(1, self.sum_values[k][1])
                    if abs(avg) > 1e-3:
                        info += ' %.4f' % avg
                    else:
                        info += ' %.4e' % avg
                else:
                    info += ' %s' % self.sum_values[k]

            self.total_width += len(info)
            if prev_total_width > self.total_width:
                info += ((prev_total_width - self.total_width) * ' ')

            sys.stdout.write(info)
            sys.stdout.flush()

            if current >= self.target:
                sys.stdout.write('\n')

        if self.verbose == 2:
            if current >= self.target:
                info = '%ds' % (now - self.start)
                for k in self.unique_values:
                    info += ' - %s:' % k
                    avg = self.sum_values[k][0] / max(1, self.sum_values[k][1])
                    if avg > 1e-3:
                        info += ' %.4f' % avg
                    else:
                        info += ' %.4e' % avg
                sys.stdout.write(info + "\n")

        self.last_update = now

    def add(self, n, values=None):
        self.update(self.seen_so_far + n, values)



import numpy as np
import spacy
from .common import remove_repeated_elements


spacy_dictionary = spacy.load('en_core_web_sm')


# parser = English()
#
#
# def get_pos_tag(sentence):
#    """ add part of speech tags for each string in sentence
#    args:
#        takes input as sentence
#    returns:
#        list containing strings with part of speech tags
#    """
#    parseText = parser(sentence)
#    pos_tag_list = []
#    for word in parseText:
#        pos_tag_list.append(word)
#        pos_tag_list.append(word.pos_)
#    return pos_tag_list
#
#
# def find_lemma(sentence):
#    """ add lemma for each string in sentence
#    args:
#        takes input as sentence
#    returns:
#        list containing strings with lemma of each string
#    """
#    parseText = parser(sentence)
#    lemma_list = []
#    for word in parseText:
#        lemma_list.append(word)
#        lemma_list.append(word.lemma_)
#    return lemma_list
#
#
# def get_lowercase_words(token):
#    """ convert token into lowercase
#    args:
#        takes input as tokens
#    returns:
#        token in lowercase
#    """
#    return token.lower_
#

def get_vector(word):
    """ get glove vector word representation
    args:
        word: string
    returns:
        300 dimensional numpy array
    """
    return spacy_dictionary(word).vector


def get_vectors(tokens, mode='glove'):
    """ get glove vector representations of all tokens
    args:
        tokens: list of strings
    returns:
        list of of 300 dimensional numpy arrays
    """
    if mode == 'glove':
        return [get_vector(token) for token in tokens]


def calculate_norm(vector):
    """ calculate norm of a vector
    args:
        vector: single dimensional numpy array
    returns:
        float norm value
    """
    return np.linalg.norm(vector)


def calculate_cosine_similarity(vector_1, vector_2):
    """ calculate cosine similarity between two vectors
    args:
        vector_1: numpy vector
        vector_2: numpy vector
    returns:
        float indicating cosine similarity
    """
    norm_1 = calculate_norm(vector_1)
    norm_2 = calculate_norm(vector_2)
    if norm_1 == 0 or norm_2 == 0:
        return 0.0
    return np.dot(vector_1, vector_2) / (norm_1 * norm_2)


def get_related_words(token, max_num=3):
    """ use glove embeddings to get close embedded words
    args:
        token: string
        max_num: maximum number of words to be returned
    returns: list of strings
    """
    spacy_token = spacy_dictionary(token)
    queries = [vocab for vocab in spacy_token.vocab if vocab.prob >= -15]
    queries = sorted(queries, key=lambda w: spacy_token.similarity(w),
                     reverse=True)
    queries = [query.lower_ for query in queries if query.lower_ != token]
    queries = remove_repeated_elements(queries)
    if len(queries) > max_num:
        queries = queries[:max_num]
    return queries



from __future__ import absolute_import
import os

from . import backend
from . import datasets
# from . import metrics
# from . import models
from . import utils

__version__ = '0.0.9'

# Obtain luvina base dir path: either ~/.luvina or /tmp.
_luvina_base_dir = os.path.expanduser('~')
if not os.access(_luvina_base_dir, os.W_OK):
    _luvina_base_dir = '/tmp'
_luvina_dir = os.path.join(_luvina_base_dir, '.luvina')

if not os.path.exists(_luvina_dir):
    try:
        os.makedirs(_luvina_dir)
    except OSError:
        # Except permission denied and potential race conditions
        # in multi-threaded environments.
        pass





def calculate_synonyms_similarities(synonyms):
    similarities = []
    for word_1, word_2 in synonyms.items():
        if len(word_2) == 0:
            continue
        elif len(word_2) >= 1:
            selected_word_2 = word_2[0]
        vector_1 = get_vector(word_1)
        vector_2 = get_vector(selected_word_2)
        if word_1 == selected_word_2:
            similarity = 1.
            print(word_1)
        elif calculate_norm(vector_1) == 0 or calculate_norm(vector_2) == 0:
            continue
        else:
            similarity = calculate_cosine_similarity(vector_1, vector_2)
        similarities.append(similarity)
    return similarities


def find_synonyms(reference_sentence, hypothesis_sentence, return_tokens=True):
    """
    Inputs: two lists of tokenized string sentences
    Returns: A dictionary with keys being the word in the reference_sentence
    and the values a list of the arguments where there is a synonym in the
    hypothesis_sentence.
    """
    synonyms_connections = OrderedDict()
    for reference_token in reference_sentence:
        reference_synonyms = list(set(get_synonyms(reference_token)))
        reference_synonyms = reference_synonyms + [reference_token]
        hypothesis_args = []
        for hypothesis_arg, hypothesis_token in enumerate(hypothesis_sentence):
            for reference_synonym in reference_synonyms:
                words_are_synonyms = hypothesis_token == reference_synonym
                words_are_the_same = hypothesis_token == reference_token
                if words_are_the_same:
                    if return_tokens:
                        hypothesis_args.append(hypothesis_token)
                    else:
                        hypothesis_args.append(hypothesis_arg)
                    break
                elif words_are_synonyms:
                    if return_tokens:
                        hypothesis_args.append(hypothesis_token)
                    else:
                        hypothesis_args.append(hypothesis_arg)

        synonyms_connections[reference_token] = hypothesis_args
    return synonyms_connections


def find_similarities(reference_sentence, hypothesis_sentence,
                      return_tokens=True):
    """
    Inputs: two lists of tokenized string sentences
    Returns: A dictionary with keys being the word in the reference_sentence
    and the values a list of the arguments where there is a synonym in the
    hypothesis_sentence.
    """
    synonyms_connections = OrderedDict()
    for reference_token in reference_sentence:
        reference_synonyms = list(set(get_synonyms(reference_token)))
        reference_synonyms = reference_synonyms + [reference_token]
        hypothesis_args = []
        for hypothesis_arg, hypothesis_token in enumerate(hypothesis_sentence):
            for reference_synonym in reference_synonyms:
                words_are_synonyms = hypothesis_token == reference_synonym
                words_are_the_same = hypothesis_token == reference_token
                if words_are_the_same:
                    if return_tokens:
                        hypothesis_args.append(hypothesis_token)
                    else:
                        hypothesis_args.append(hypothesis_arg)
                    break
                elif words_are_synonyms:
                    if return_tokens:
                        hypothesis_args.append(hypothesis_token)
                    else:
                        hypothesis_args.append(hypothesis_arg)

        synonyms_connections[reference_token] = hypothesis_args
    return synonyms_connections


def edit(hypothesis_sentence, token_connections,
         start_string='**', end_string='**'):
    edited_hypothesis = []
    hypothesis_highlights = list(chain(*list(token_connections.values())))
    for hypothesis_token in hypothesis_sentence:
        if hypothesis_token in hypothesis_highlights:
            hypothesis_token = start_string + hypothesis_token + end_string
        edited_hypothesis.append(hypothesis_token)
    edited_hypothesis = ' '.join(edited_hypothesis)
    return edited_hypothesis





import re
from collections import OrderedDict

from nltk.tokenize import word_tokenize
from nltk.corpus import wordnet
from nltk.corpus import stopwords
from nltk.stem import WordNetLemmatizer
from nltk import download
from nltk import pos_tag
from nltk.metrics import edit_distance
from nltk.util import ngrams

# -------------------------------------------------------------------
# NLTK GLOBAL VARIABLES
# -------------------------------------------------------------------
lemmatizer = WordNetLemmatizer()
replacement_patterns = [
        (r'won\'t', 'will not'),
        (r'can\'t', 'cannot'),
        (r'i\'m', 'i am'),
        (r'ain\'t', 'is not'),
        (r'(\w+)\'ll', '\g<1> will'),
        (r'(\w+)n\'t', '\g<1> not'),
        (r'(\w+)\'ve', '\g<1> have'),
        (r'(\w+)\'s', '\g<1> is'),
        (r'(\w+)\'re', '\g<1> are'),
        (r'(\w+)\'d', '\g<1> would')
]

patterns = []
for regular_expression, replacement in replacement_patterns:
    patterns.append((re.compile(regular_expression), replacement))

repeat_regular_expresssion = re.compile(r'(\w*)(\w)\2(\w*)')
repeat_replacement = r'\1\2\3'
# -------------------------------------------------------------------


def download_nltk_data(package_name='all'):
    """ download necessary data from NLTK
    args:
        package_name: string containing the package name to install
    returns:
        None
    """
    if package_name is 'all':
        data = ['punkt', 'wordnet', 'stopwords', 'averaged_perceptron_tagger']
        for package in data:
            download(package)
    else:
        download(package)


def tokenize(sentence, lowercase=True):
    """ tokenize a setence
    args:
        sentence: string
        lowercase: boolean flag for returning lowercase format
    returns:
        a list of the tokens in the sentence
    """
    tokens = word_tokenize(sentence)
    if lowercase:
        tokens = [token.lower() for token in tokens]
    return tokens


def _remove_repeated_elements(tokens):
    """ removes repeated tokens
    args:
        tokens: a list of tokens
    returns:
        filtered_tokens: a list of tokens without repeated tokens
    """
    filtered_tokens = list(OrderedDict((token, None) for token in tokens))
    return filtered_tokens


def get_synonyms(token):
    """ get synonyms of word using wordnet
    args:
        token: string
    returns:
        synonyms: list containing synonyms as strings
    """
    synonyms = []
    if len(wordnet.synsets(token)) == 0:
        return None
    for synset in wordnet.synsets(token):
        for lemma in synset.lemmas():
            synonyms.append(lemma.name())
    synonyms = _remove_repeated_elements(synonyms)
    synonyms.remove(token)
    return synonyms


def get_definition(token):
    """ get definition of word using wordnet
    args:
        token: string
    returns:
        definition: string containing definition of word
    """
    synsets = wordnet.synsets(token)
    if len(synsets) == 0:
        return []
    return synsets[0].definition()


def get_stop_words(language='english'):
    """ get English stop words
    args:
        language: language for which the stop words will be used
    returns:
        stop_words= list of stop words in string format.
    """
    stop_words = list(set(stopwords.words(language)))
    return stop_words


def filter_tokens(tokens, filters):
    """ eliminates tokens contained in filters
    args:
        tokens: list of tokens to be filtered.
        filters: list of tokens that will get removed from 'tokens'.
    returns:
        filtered_tokens: a list of tokens that don't contain any
        token from filters.
    """
    filtered_tokens = []
    for token in tokens:
        if token not in filters:
            filtered_tokens.append(token)
    return filtered_tokens


def calculate_wordnet_similarity(token_1, token_2):
    """ calculates similarity between two tokens using wordnet
    args:
        token_1: string
        token_2: string
    returns:
        wordnet_similarity: int score for wordnet similarity or None
        if either given token is not a word.
    """
    synset_1 = wordnet.synsets(token_1)
    synset_2 = wordnet.synsets(token_2)
    if len(synset_1) == 0 or len(synset_2) == 0:
        return None
    wordnet_similarity = synset_1[0].wup_similarity(synset_2[0])
    return wordnet_similarity


def lemmatize(token, pos='n'):
    """ lemmatize a token using a selected part of speech
    args:
        token: string
        pos = parth_of_speech of the given token
    returns
        lemmatized_token: a lemmatized/reduced token
    """
    lemmatized_token = lemmatizer.lemmatize(token, pos=pos)
    return lemmatized_token


def expand_contractions(sentence):
    """ expand English contractions found in sentence/token:
    args:
        sentence: string of complete sentence or a single token
    returns:
        sentence = string sentence/token without contractions
    """
    for (pattern, replacement) in patterns:
        sentence = re.sub(pattern, replacement, sentence)
    return sentence


def remove_repeated_characters(token):
    """ removes recursively repeated characters in a token if token not
    found in dictionary
    args:
        token: string
    returns:
        replaced_token: string
    """
    if wordnet.synsets(token):
        return token
    replaced_token = repeat_regular_expresssion.sub(repeat_replacement, token)
    if replaced_token != token:
        return remove_repeated_characters(replaced_token)
    else:
        return replaced_token


def tag_pos(tokens):
    """ add parts of speech tags for every token given
    args:
        tokens: list of strings
    returns
        list of lists containing string token and string part of speech
    """
    return pos_tag(tokens)


def calculate_levenshtein_distance(token_1, token_2):
    """ get number of characters that need to be substituted,
    inserted or deleted to go from token_1 to token_2.
    args:
        token_1: string
        token_2: string
    returns:
        int
    """
    return edit_distance(token_1, token_2)


def make_ngrams(sequence, n, pad_left=False, pad_right=False,
                left_pad_symbol=None, right_pad_symbol=None):
    """ return the ngrams made by sequence of items
    args:
        tokens: list of strings/tokens
        n: number of subsequent items
    returns:
        list of lists
    """
    return list(ngrams(sequence, n, pad_left, pad_right,
                       left_pad_symbol, right_pad_symbol))



from __future__ import absolute_import
from collections import Counter
from collections import OrderedDict
from itertools import chain
import numpy as np
from .enchant_backend import in_dictionary
from .enchant_backend import suggest_words
from .nltk_backend import make_ngrams
from .nltk_backend import calculate_levenshtein_distance


def correct_misspelling(token, distance_threshold=2):
    """ correct misspelling of token by comparing suggestions and measuring
    the levenshtein distance
    args:
        token: string
        distance_threshold: int that describes the maximum number of desired
        changed characters from 'token'
    returns:
        best_suggestion: string containing suggestion
        token: original token is not match was found
    """
    if in_dictionary(token):
        return token
    suggested_words = suggest_words(token)
    if suggested_words is not None:
        num_modified_characters = []
        for suggested_word in suggested_words:
            distance = calculate_levenshtein_distance(token, suggested_word)
            num_modified_characters.append(distance)
        # this min is showing errors since it takes an empy/none variable as inputen    
        min_num_modified_characters = min(num_modified_characters)
        best_arg = num_modified_characters.index(min_num_modified_characters)
        if distance_threshold > min_num_modified_characters:
            best_suggestion = suggested_words[best_arg]
            return best_suggestion
        else:
            return token
    else:
        return token


def remove_repeated_elements(tokens):
    """ removes repeated tokens
    args:
        tokens: a list of tokens
    returns:
        filtered_tokens: a list of tokens without repeated tokens
    """
    filtered_tokens = list(OrderedDict((token, None) for token in tokens))
    return filtered_tokens


def join(tokens):
    """ construct a string sentence from joining tokens with spaces
    args:
        tokens: list of strings
    returns:
        joined_tokens: a string containing all tokens
    """
    joined_tokens = ' '.join(tokens)
    return joined_tokens


def compute_jaccard_similarity(query, document):
    """ function taken explicitly from:
    http://billchambers.me/tutorials/2014/12/21/tf-idf-explained-in-python.html
    calculates the intersection over union of a query
    in a given document.
    """
    intersection = set(query).intersection(set(document))
    union = set(query).union(set(document))
    return len(intersection)/len(union)


def calculate_jaccard_coefficient(a, b):
    union = list(set(a + b))
    intersection = list(set(a) - (set(a) - set(b)))
    jaccard_coefficient = float(len(intersection)) / len(union)
    return jaccard_coefficient


def correct_misspelling_ngram(token, levenshtein_treshold=3):
    """ corrects token by suggesting words and filtering them
    using the levenhstein distance. Then it takes all filtered
    words and chooses the one with the highest jaccard coefficient
    calculated using bigrams.
    args:
        token: string
        levenshtein threshold: int
    returns:
        token: string
    """
    if in_dictionary(token):
        return token
    suggested_words = suggest_words(token)
    jaccard_coefficients = []
    best_suggested_words = []
    if suggested_words is not None:
        token_bigrams = make_ngrams(token, 2)
        for suggested_word in suggested_words:
            distance = calculate_levenshtein_distance(token, suggested_word)
            if distance < levenshtein_treshold:
                suggested_bigrams = make_ngrams(suggested_word, 2)
                jaccard_coefficient = calculate_jaccard_coefficient(
                                    token_bigrams, suggested_bigrams)
                jaccard_coefficients.append(jaccard_coefficient)
                best_suggested_words.append(suggested_word)
        highest_jaccard = max(jaccard_coefficients)
        best_arg = jaccard_coefficients.index(highest_jaccard)
        word = best_suggested_words[best_arg]
        return word
    else:
        return word


def get_token_frequencies(sentences):
    """ count the number of times all tokens appear in all sentences.
    args:
        sentences: list of sentences where each sentence contains a
        list of tokens
    returns:
        word_frequencies: list containing two elements lists with the
        word (string) and an integer describing the amount of times
        the word has appeared in the all the sentences.
    """
    word_frequencies = Counter(chain(*sentences)).most_common()
    return word_frequencies


def pad(tokens, max_token_size=22, remove=False, BOS_token='<BOS>',
        EOS_token='<EOS>', PAD_token='<PAD>'):
    """
    args:
        tokens: a list of strings containing tokens
        max_token_size: Max token size including the EOS and BOS tokens.
        remove: Boolean flag for determining if it should remove tokens
        bigger than max_token_size.
        BOS_token: string beginning of the sentence token.
        EOS_token: string end of the sentence token.
        PAD_token: string pad token.
    returns:
        tuple of strings containing all added/removed tokens.
        Tokens will get removed if the remove flag is enabled
        and the number of tokens is bigger than max_token_size.
    """

    sentence = list(tokens)
    max_sentence_length = max_token_size - 2
    if len(sentence) == max_sentence_length:
        padded_sentence = [BOS_token] + sentence + [EOS_token]
    elif len(sentence) > max_sentence_length and not remove:
        sentence = sentence[:max_sentence_length]
        padded_sentence = [BOS_token] + sentence + [EOS_token]
    elif len(sentence) > max_sentence_length and remove:
        return []
    elif len(sentence) < max_sentence_length:
        padded_sentence = [BOS_token] + sentence + [EOS_token]
        pad_size = max_token_size - len(padded_sentence)
        pad = [PAD_token] * pad_size
        padded_sentence = padded_sentence + pad
    return padded_sentence


def remove_infrequent_tokens(sentences, word_frequencies, min_frequency=3):
    pass
#######################################################
# Deprecated functions
#######################################################

# def remove_long_sentences(sentences, max_token_size=25):
#    # TODO: Add y predictions that should/could also be filtered
#    # you pass y and iterate over its first dimension.
#    """ removes sentences with a length bigger that max_length.
#    args:
#        sentences: list of lists containing strings/tokens.
#        max_length: int > 0
#    returns:
#        filtered_sentences: list pf lists containing a strings/tokens.
#    """
#    filtered_sentences = []
#    for tokens in sentences:
#        if len(tokens) <= max_token_size:
#            filtered_sentences.append(tokens)
#    return filtered_sentences
#
#
# def remove_long_sentences(sentences, associated_data=None,
#    max_token_size=25):
#    # TODO: Add y predictions that should/could also be filtered
#    # you pass y and iterate over its first dimension.
#    """ removes sentences with a length bigger that max_length.
#    args:
#        sentences: list of lists containing strings/tokens.
#        data: Additional associated data that should be removed
#        if a sentence is removed e. g. labels or another pair of sentences
#        max_length: int > 0
#    returns:
#        filtered_sentences: list pf lists containing a strings/tokens.
#    """
#    mask = get_token_size_mask(sentences, max_token_size)
#    sentences = np.asarray(sentences)[mask]
#    associated_data = np.asarray(associated_data)[mask]

def get_token_size_mask(sentences, max_token_size=25):
    """ returns mask containing True for sentences with less
    or equal amount of tokens than the max_token_size, and
    false otherwise.
    args:
        sentences: list of lists containing strings/tokens
        max_length: int > 0
    returns:
        boolean numpy array of size(sentences)
    """
    mask = np.zeros(shape=len(sentences))
    for sentence_arg, tokens in enumerate(sentences):
        if len(tokens) <= max_token_size:
            mask[sentence_arg] = True
        else:
            mask[sentence_arg] = False
    return mask


def mask_data(data, mask):
    if data is not np.ndarray:
        data = np.asarray(data)
        return data[mask].tolist()
    else:
        return data[mask]


def pad_with_zeros(sentences, max_length=25):
    data = []
    for vectors in sentences:
        vectors = np.asarray(vectors)
        sentence_length, embedding_dimension = vectors.shape
        missing_zeros = max_length - sentence_length
        zero_array = np.zeros(shape=(missing_zeros, embedding_dimension))
        vectors = np.concatenate((vectors, zero_array), axis=0)
        data.append(vectors.tolist())
    return data



from keras.models import Model
from keras.layers import Input, Masking
from keras.layers.recurrent import GRU
from keras.layers.merge import Dot
from keras.layers.wrappers import Bidirectional


def SiameseLSTM(max_token_length, hidden_size, embedding_size=300):
    text_input_1 = Input(shape=(max_token_length, embedding_size),
                         name='text_1')
    text_mask_1 = Masking(mask_value=0.0, name='text_mask_1')(text_input_1)
    # text_dropout_1 = Dropout(.5, name='text_dropout_1')(text_mask_1)

    text_input_2 = Input(shape=(max_token_length, embedding_size),
                         name='text_2')
    text_mask_2 = Masking(mask_value=0.0, name='text_mask_2')(text_input_2)
    # text_dropout_2 = Dropout(.5, name='text_dropout_2')(text_mask_2)

    lstm_1_a = Bidirectional(GRU(units=hidden_size,
                                 return_sequences=True,
                                 name='RNN_1_a'))(text_mask_1)

    lstm_1_b = Bidirectional(GRU(units=hidden_size,
                                 return_sequences=False,
                                 name='RNN_1_b'))(lstm_1_a)

    """
    lstm_1_c = Bidirectional(GRU(units=hidden_size,
                                 return_sequences=False,
                                 name='RNN_1_c'))(lstm_1_b)
    """

    lstm_2_a = Bidirectional(GRU(units=hidden_size,
                                 return_sequences=True,
                                 name='RNN_2_a'))(text_mask_2)

    lstm_2_b = Bidirectional(GRU(units=hidden_size,
                                 return_sequences=False,
                                 name='RNN_2_b'))(lstm_2_a)

    """
    lstm_2_c = Bidirectional(GRU(units=hidden_size,
                                 return_sequences=False,
                                 name='RNN_2_c'))(lstm_2_b)
    """

    cosine_similarity = Dot(axes=1, normalize=True,
                            name='cosine_similarity')([lstm_1_b, lstm_2_b])

    model = Model(inputs=[text_input_1, text_input_2],
                  outputs=cosine_similarity)

    return model


if __name__ == '__main__':
    model = SiameseLSTM(25, 50)
    model.summary()



from luvina.utils.data_utils import get_file
import pandas as pd
import glob
from os.path import basename
import os
import numpy as np


def _get_string_data(root_path):
    # data_path = '../datasets/short_answer_grading_v2/raw/'
    data_path = root_path + '/raw/'
    questions_filename = data_path + 'questions'
    teacher_answers_filename = data_path + 'answers'
    student_answers_filename = data_path + 'all'
    questions = _read_file(questions_filename)
    teacher_answers = _read_file(teacher_answers_filename)
    student_answers = _read_file(student_answers_filename)
    return questions, teacher_answers, student_answers


def _read_file(filename):
    data_file = open(filename, 'r', encoding='latin-1')
    texts = []
    data_ids = []
    for line in data_file:
        if line[3] == ' ':
            data_id = line[:3]
            text = line[4:-2]
        elif line[4] == ' ':
            data_id = line[:4]
            text = line[5:-2]
        elif line[5] == ' ':
            data_id = line[:5]
            text = line[6:-2]
        else:
            raise NotImplementedError('Space value not found')
        data_ids.append(data_id)
        texts.append(text)
    data_frame = pd.DataFrame({'id': data_ids, 'text': texts})
    return data_frame


def _get_scores(root_path):
    data_path = root_path + '/scores/'
    score_paths = glob.glob(data_path + '*')
    scores = dict()
    for scores_path in score_paths:
        score_files = glob.glob(scores_path + '/*')
        question_scores = []
        for score_file in score_files:
            file_scores = []
            _score_file = open(score_file, 'r')
            for line in _score_file:
                file_scores.append(float(line))
            question_scores.append(file_scores)
            question_scores_array = np.asarray(question_scores)
            if np.max(question_scores) > 5.0:
                question_scores = (question_scores_array / 2.).tolist()
        scores_key = basename(scores_path)
        scores[scores_key] = question_scores
    return scores


def _convert_to_dictionary(string_data):
    data = dict()
    for arg, key in enumerate(string_data.id.tolist()):
        data[key] = string_data.text.iloc[arg]
    return data


def get_data(root_path):
    questions, teacher_answers, student_answers = _get_string_data(root_path)
    scores = _get_scores(root_path)

    teacher_answers = _convert_to_dictionary(teacher_answers)
    questions = _convert_to_dictionary(questions)

    teacher_answers_list = []
    student_answers_list = []
    questions_list = []
    scores_list = []
    keys_list = []
    for key in student_answers.id.unique().tolist():
        masked_student_answers = student_answers[student_answers.id == key]
        masked_scores = scores[key]
        for score_set in masked_scores:
            for mask_arg in range(len(masked_student_answers.text)):
                student_answer = masked_student_answers.text.iloc[mask_arg]
                student_answers_list.append(student_answer)
                scores_list.append(score_set[mask_arg])
                teacher_answers_list.append(teacher_answers[key])
                questions_list.append(questions[key])
                keys_list.append(key)

    return pd.DataFrame({'id': keys_list,
                         'questions': questions_list,
                         'teacher': teacher_answers_list,
                         'student': student_answers_list,
                         'scores': scores_list})


def load_data(path='ShortAnswerGrading_v2.0'):
    """Loads the Short Answer grading data.

    # Arguments
        path: path where to cache the dataset locally
            (relative to ~/.keras/datasets).

    # Returns
        Tuple of Numpy arrays: `(x_train, y_train), (x_test, y_test)`.
    """
    origin = ('http://web.eecs.umich.edu/~mihalcea/' +
              'downloads/ShortAnswerGrading_v2.0.zip')
    root_path = get_file(path, origin=origin, extract=True,
                         cache_subdir='datasets/short_answer_grading')
    root_path = os.path.dirname(root_path) + '/data'
    return get_data(root_path)


data = load_data()



""" Data utils taken explicitly from keras.
Utilities for file download and caching.
"""
from __future__ import absolute_import
from __future__ import print_function

import hashlib
import os
import shutil
import tarfile
import zipfile

import six
from six.moves.urllib.error import HTTPError
from six.moves.urllib.error import URLError
from six.moves.urllib.request import urlretrieve


from ..utils.generic_utils import Progbar


def _extract_archive(file_path, path='.', archive_format='auto'):
    """Extracts an archive if it matches tar, tar.gz, tar.bz, or zip formats.

    # Arguments
        file_path: path to the archive file
        path: path to extract the archive file
        archive_format: Archive format to try for extracting the file.
            Options are 'auto', 'tar', 'zip', and None.
            'tar' includes tar, tar.gz, and tar.bz files.
            The default 'auto' is ['tar', 'zip'].
            None or an empty list will return no matches found.

    # Returns
        True if a match was found and an archive extraction was completed,
        False otherwise.
    """
    if archive_format is None:
        return False
    if archive_format is 'auto':
        archive_format = ['tar', 'zip']
    if isinstance(archive_format, six.string_types):
        archive_format = [archive_format]

    for archive_type in archive_format:
        if archive_type is 'tar':
            open_fn = tarfile.open
            is_match_fn = tarfile.is_tarfile
        if archive_type is 'zip':
            open_fn = zipfile.ZipFile
            is_match_fn = zipfile.is_zipfile

        if is_match_fn(file_path):
            with open_fn(file_path) as archive:
                try:
                    archive.extractall(path)
                except (tarfile.TarError, RuntimeError,
                        KeyboardInterrupt):
                    if os.path.exists(path):
                        if os.path.isfile(path):
                            os.remove(path)
                        else:
                            shutil.rmtree(path)
                    raise
            return True
    return False


def get_file(fname,
             origin,
             untar=False,
             md5_hash=None,
             file_hash=None,
             cache_subdir='datasets',
             hash_algorithm='auto',
             extract=False,
             archive_format='auto',
             cache_dir=None):
    """Downloads a file from a URL if it not already in the cache.

    By default the file at the url `origin` is downloaded to the
    cache_dir `~/.keras`, placed in the cache_subdir `datasets`,
    and given the filename `fname`. The final location of a file
    `example.txt` would therefore be `~/.keras/datasets/example.txt`.

    Files in tar, tar.gz, tar.bz, and zip formats can also be extracted.
    Passing a hash will verify the file after download. The command line
    programs `shasum` and `sha256sum` can compute the hash.

    # Arguments
        fname: Name of the file. If an absolute path `/path/to/file.txt` is
            specified the file will be saved at that location.
        origin: Original URL of the file.
        untar: Deprecated in favor of 'extract'.
            boolean, whether the file should be decompressed
        md5_hash: Deprecated in favor of 'file_hash'.
            md5 hash of the file for verification
        file_hash: The expected hash string of the file after download.
            The sha256 and md5 hash algorithms are both supported.
        cache_subdir: Subdirectory under the Keras cache dir where the file is
            saved. If an absolute path `/path/to/folder` is
            specified the file will be saved at that location.
        hash_algorithm: Select the hash algorithm to verify the file.
            options are 'md5', 'sha256', and 'auto'.
            The default 'auto' detects the hash algorithm in use.
        extract: True tries extracting the file as an Archive, like tar or zip.
        archive_format: Archive format to try for extracting the file.
            Options are 'auto', 'tar', 'zip', and None.
            'tar' includes tar, tar.gz, and tar.bz files.
            The default 'auto' is ['tar', 'zip'].
            None or an empty list will return no matches found.
        cache_dir: Location to store cached files, when None it
            defaults to the [Keras Directory]
            (/faq/#where-is-the-keras-configuration-filed-stored).

    # Returns
        Path to the downloaded file
    """
    if cache_dir is None:
        cache_dir = os.path.expanduser(os.path.join('~', '.luvina'))
    if md5_hash is not None and file_hash is None:
        file_hash = md5_hash
        hash_algorithm = 'md5'
    datadir_base = os.path.expanduser(cache_dir)
    if not os.access(datadir_base, os.W_OK):
        datadir_base = os.path.join('/tmp', '.luvina')
    datadir = os.path.join(datadir_base, cache_subdir)
    if not os.path.exists(datadir):
        os.makedirs(datadir)

    if untar:
        untar_fpath = os.path.join(datadir, fname)
        fpath = untar_fpath + '.tar.gz'
    else:
        fpath = os.path.join(datadir, fname)

    download = False
    if os.path.exists(fpath):
        # File found; verify integrity if a hash was provided.
        if file_hash is not None:
            if not validate_file(fpath, file_hash, algorithm=hash_algorithm):
                print('A local file was found, but it seems to be '
                      'incomplete or outdated because the ' + hash_algorithm +
                      ' file hash does not match the original value of ' +
                      file_hash + ' so we will re-download the data.')
                download = True
    else:
        download = True

    if download:
        print('Downloading data from', origin)

        class ProgressTracker(object):
            # Maintain progbar for the lifetime of download.
            # This design was chosen for Python 2.7 compatibility.
            progbar = None

        def dl_progress(count, block_size, total_size):
            if ProgressTracker.progbar is None:
                if total_size is -1:
                    total_size = None
                ProgressTracker.progbar = Progbar(total_size)
            else:
                ProgressTracker.progbar.update(count * block_size)

        error_msg = 'URL fetch failure on {}: {} -- {}'
        try:
            try:
                urlretrieve(origin, fpath, dl_progress)
            except URLError as e:
                raise Exception(error_msg.format(origin, e.errno, e.reason))
            except HTTPError as e:
                raise Exception(error_msg.format(origin, e.code, e.msg))
        except (Exception, KeyboardInterrupt) as e:
            if os.path.exists(fpath):
                os.remove(fpath)
            raise
        ProgressTracker.progbar = None

    if untar:
        if not os.path.exists(untar_fpath):
            _extract_archive(fpath, datadir, archive_format='tar')
        return untar_fpath

    if extract:
        _extract_archive(fpath, datadir, archive_format)

    return fpath


def _hash_file(fpath, algorithm='sha256', chunk_size=65535):
    """Calculates a file sha256 or md5 hash.

    # Example

    ```python
       >>> from keras.data_utils import _hash_file
       >>> _hash_file('/path/to/file.zip')
       'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
    ```

    # Arguments
        fpath: path to the file being validated
        algorithm: hash algorithm, one of 'auto', 'sha256', or 'md5'.
            The default 'auto' detects the hash algorithm in use.
        chunk_size: Bytes to read at a time, important for large files.

    # Returns
        The file hash
    """
    if (algorithm is 'sha256') or (algorithm is 'auto' and len(hash) is 64):
        hasher = hashlib.sha256()
    else:
        hasher = hashlib.md5()

    with open(fpath, 'rb') as fpath_file:
        for chunk in iter(lambda: fpath_file.read(chunk_size), b''):
            hasher.update(chunk)

    return hasher.hexdigest()


def validate_file(fpath, file_hash, algorithm='auto', chunk_size=65535):
    """Validates a file against a sha256 or md5 hash.

    # Arguments
        fpath: path to the file being validated
        file_hash:  The expected hash string of the file.
            The sha256 and md5 hash algorithms are both supported.
        algorithm: Hash algorithm, one of 'auto', 'sha256', or 'md5'.
            The default 'auto' detects the hash algorithm in use.
        chunk_size: Bytes to read at a time, important for large files.

    # Returns
        Whether the file is valid
    """
    if ((algorithm is 'sha256') or
            (algorithm is 'auto' and len(file_hash) is 64)):
        hasher = 'sha256'
    else:
        hasher = 'md5'

    if str(_hash_file(fpath, hasher, chunk_size)) == str(file_hash):
        return True
    else:
        return False



import keras.backend as K


def calculate_r2_score(y_true, y_pred):
    """ calculates the R2 score
    args:
        y_true: keras tensor
        y_pred: keras tensor
    returns:
        R2: tensor
    """
    SS_res = K.sum(K.square(y_true - y_pred))
    SS_tot = K.sum(K.square(y_true - K.mean(y_true)))
    R2 = 1 - (SS_res / (SS_tot + K.epsilon()))
    return R2



import pandas as pd
from luvina.utils.data_utils import get_file


origin = 'https://raw.githubusercontent.com/alvations/stasis/master/sts.csv'
file_name = 'semantic_text_similarity.csv'


def load_data():
    filepath = get_file(file_name, origin)
    data = pd.read_csv(filepath, sep='\t', error_bad_lines=False,
                       warn_bad_lines=False)
    data = data.dropna()
    return data



from setuptools import setup
from setuptools import find_packages

setup(name='luvina',
      version='0.0.24',
      description='High-level API for Natural Language Processing in NLTK',
      author='Octavio Arriaga',
      author_email='arriaga.camargo@gmail.com',
      url='https://github.com/oarriaga/luvina',
      license='MIT',
      install_requires=['nltk', 'pyenchant', 'spacy'],
      packages=find_packages()
      )
